-- some basic options
local vim = vim

-- nvim tree said to disable netrw
vim.g.loaded_netrw = 1
vim.g.loaded_netrwPlugin = 1

vim.opt.tabstop = 4
vim.opt.shiftwidth = 4
vim.opt.softtabstop = 2
vim.opt.expandtab = true
vim.opt.termguicolors = true
vim.opt.autoindent = true
vim.opt.smartindent = true
vim.opt.smarttab = true
vim.opt.number = true
vim.o.hidden = true
vim.o.signcolumn = 'auto'
-- vim.o.breakindent = false -- I think this looks bad
vim.o.errorbells = true
vim.o.clipboard = "unnamedplus"
vim.g.noswapfile = true
vim.cmd [[set mouse=a]] -- enable mouse
vim.wo.cursorline = true
vim.opt.hlsearch = false
vim.opt.wrap = true
vim.opt.linebreak = true
vim.opt.showmatch = true
vim.opt.completeopt = { 'menu', 'menuone', 'noselect', 'noinsert' }
vim.opt.shortmess = vim.opt.shortmess + "c"

vim.o.laststatus = 3
vim.o.cmdheight = 0 -- hides commandbar when not in use, looks ok

vim.opt.spelllang = 'en_us'

vim.g.python3_host_prog = '/home/frosty/.dotfiles/nvim/env/bin/python'

vim.api.nvim_create_user_command('W', 'write', {})

vim.opt.foldmethod = "expr"
vim.opt.foldexpr = "nvim_treesitter#foldexpr()"
-- hit zc to enable folding
vim.o.foldenable = false

vim.g.neomake_open_list = 2

function map(mode, shortcut, command)
    vim.api.nvim_set_keymap(mode, shortcut, command, { noremap = true, silent = true })
end

-- normal mode map
function nmap(shortcut, command)
    map('n', shortcut, command)
end

-- insert mode map
function imap(shortcut, command)
    map('i', shortcut, command)
end

vim.opt.guifont = "Hack Nerd Font:h12"
if vim.g.neovide then
    vim.g.neovide_refresh_rate = 140
    -- change to whatever font you prefer
    -- can check font with fc-cache on linux
    vim.g.neovide_hide_mouse_when_typing = true
    vim.g.neovide_refresh_rate_idle = 140
    vim.g.neovide_scale_factor = 1.0
    -- zoom with CTRL - and CTRL +
    local change_scale_factor = function(delta)
        vim.g.neovide_scale_factor = vim.g.neovide_scale_factor * delta
    end
    vim.keymap.set("n", "<C-=>", function()
        change_scale_factor(1.25)
    end)
    vim.keymap.set("n", "<C-->", function()
        change_scale_factor(1 / 1.25)
    end)
end


-- <leader> is leader key; default is \
nmap("<leader>v", "<cmd>NvimTreeToggle<cr>")
nmap("<leader>ec", "<cmd>e ~/.config/nvim/init.lua<cr>")
nmap("<leader>sc", "<cmd>source ~/.config/nvim/init.lua<cr><cmd>PackerClean<cr><cmd>PackerCompile<cr>")

-- buffer commands
nmap("<leader>b", "<cmd>BufferLineCyclePrev<cr>")
nmap("<leader>n", "<cmd>BufferLineCycleNext<cr>")
nmap("<leader>gb", "<cmd>BufferLinePick<cr>")
nmap("<leader>w", "<cmd>bdelete!<cr>")
nmap("<leader>f", "<cmd>Telescope live_grep<cr>")
-- toggle terminal in both modes
nmap("<leader>t", "<cmd>FloatermToggle<cr>")
map("t", "<leader>t", "<cmd>FloatermToggle<cr>")

nmap("<leader>dd", "<cmd> lua vim.diagnostic.open_float() <CR>")

vim.keymap.set("n", "<C-a>", function()
    require("dial.map").manipulate("increment", "normal")
end)
vim.keymap.set("n", "<C-x>", function()
    require("dial.map").manipulate("decrement", "normal")
end)

local augroup = vim.api.nvim_create_augroup("Lspformatting", {})

-- lsp on_attach mappings
local on_attach = function(client, bufnr)
    if (not bufnr) then
        return
    end
    -- Enable completion triggered by <c-x><c-o>
    vim.api.nvim_buf_set_option(bufnr, 'omnifunc', 'v:lua.vim.lsp.omnifunc')

    -- Mappings.
    -- See `:help vim.lsp.*` for documentation on any of the below functions
    local bufopts = { noremap = true, silent = true, buffer = bufnr }
    vim.keymap.set('n', 'gD', vim.lsp.buf.declaration, bufopts)
    vim.keymap.set('n', 'gd', vim.lsp.buf.definition, bufopts)
    vim.keymap.set('n', 'K', vim.lsp.buf.hover, bufopts)
    vim.keymap.set('n', 'gi', vim.lsp.buf.implementation, bufopts)
    vim.keymap.set('n', '<C-k>', vim.lsp.buf.signature_help, bufopts)
    vim.keymap.set('n', '<space>wa', vim.lsp.buf.add_workspace_folder, bufopts)
    vim.keymap.set('n', '<space>wr', vim.lsp.buf.remove_workspace_folder, bufopts)
    vim.keymap.set('n', '<space>wl', function()
        print(vim.inspect(vim.lsp.buf.list_workspace_folders()))
    end, bufopts)
    vim.keymap.set('n', '<space>D', vim.lsp.buf.type_definition, bufopts)
    vim.keymap.set('n', '<space>rn', vim.lsp.buf.rename, bufopts)
    vim.keymap.set('n', '<space>ca', vim.lsp.buf.code_action, bufopts)
    vim.keymap.set('n', 'gr', vim.lsp.buf.references, bufopts)
    vim.keymap.set('n', '<space>f', '<cmd>lua vim.lsp.buf.format({ async = true })<CR>', bufopts)

    -- require "coq".lsp_ensure_capabilities {}
    require 'illuminate'.on_attach(client)

    -- format on save does not work for python for some reason
    if client.supports_method("textDocument/formatting") then
        vim.api.nvim_clear_autocmds({ group = augroup, buffer = bufnr })
        vim.api.nvim_create_autocmd("BufWritePre", {
            group = augroup,
            buffer = bufnr,
            callback = function()
                vim.lsp.buf.format({
                    bufnr = bufnr,
                })
            end
        })
    end
end

local lsp = require "lspconfig"
local lsp_defaults = lsp.util.default_config
local capabilities = require('cmp_nvim_lsp').default_capabilities()
capabilities.offsetEncoding = { "utf-16" }

lsp.clangd.setup { on_attach = on_attach, capabilities = capabilities }
lsp.ts_ls.setup { on_attach = on_attach, capabilities = capabilities }
lsp.hls.setup { on_attach = on_attach, capabilities = capabilities } -- haskell language server
lsp.pyright.setup { on_attach = on_attach, capabilities = capabilities }
lsp.svelte.setup { on_attach = on_attach, capabilities = capabilities }
lsp.marksman.setup { on_attach = on_attach, capabilities = capabilities }
lsp.rust_analyzer.setup { on_attach = on_attach,
    capabilities = capabilities,
    settings = {
        ["rust-analyzer"] = {
            checkOnSave = true,
            check = {
                enable = true,
                command = "clippy",
                features = "all"
            }
        }
    }
}
lsp.lua_ls.setup { on_attach = on_attach,
    capabilities = capabilities,
    settings = {
        Lua = {
            diagnostics = {
                globals = { 'vim' }
            }
        }
    }
}

lsp.tailwindcss.setup { on_attach = on_attach,
    capabilities = capabilities
}

lsp.julials.setup {
    capabilities = capabilities,
    on_new_config = function(new_config, _)
        local julia = vim.fn.expand("~/.julia/environments/nvim-lspconfig/bin/julia")
        if lsp.util.path.is_file(julia) then
            vim.notify("running julials")
            new_config.cmd[1] = julia
        end
    end,
    on_attach = on_attach
}

-- kill eslint_d on exit
local on_leave_group = vim.api.nvim_create_augroup("OnLeaveGroup", {})
vim.api.nvim_create_autocmd("VimLeave", {
    group = on_leave_group,
    command = "silent !killall -q eslint_d",
})

--[[ local vertical_help = vim.api.nvim_create_augroup("vertical_help", {})
vim.api.nvim_create_autocmd("FileType", {
    pattern = 'help',
    group = vertical_help,
    command = ":windcmd L"
})
]]

-- packer boilerplate
local fn = vim.fn
local install_path = fn.stdpath('data') .. '/site/pack/packer/start/packer.nvim'
if fn.empty(fn.glob(install_path)) > 0 then
    packer_bootstrap = fn.system({ 'git', 'clone', '--depth', '1', 'https://github.com/wbthomason/packer.nvim',
        install_path })
end

vim.cmd("colorscheme iceberg")

return require('packer').startup(function(use)
    use 'wbthomason/packer.nvim'
    use 'nvim-lua/plenary.nvim'
    use 'neovim/nvim-lspconfig'
    use {
        'nvim-treesitter/nvim-treesitter',
        run = ':TSUpdate',
        config = function()
            require('nvim-treesitter.configs').setup {
                highlight = {
                    enable = true
                },
            }
        end
    }
    use {
        'echasnovski/mini.icons'
    }
    use {
        'nvim-tree/nvim-web-devicons',
    }
    use {
        'nvim-tree/nvim-tree.lua',
        after = "nvim-web-devicons",
        requires = "nvim-tree/nvim-web-devicons",
        config = function()
            require('nvim-tree').setup {
                hijack_cursor = true,
                disable_netrw = true,
                select_prompts = true,
                renderer = {
                    add_trailing = true,
                    hidden_display = "all",
                    highlight_opened_files = "icon"
                }
            }
        end
    }
    use {
        "hrsh7th/nvim-cmp",
        requires = {
            "hrsh7th/cmp-buffer", "hrsh7th/cmp-nvim-lsp",
            "L3MON4D3/LuaSnip", "rafamadriz/friendly-snippets",
            'hrsh7th/cmp-nvim-lua', 'hrsh7th/cmp-path',
            'hrsh7th/cmp-calc', 'hrsh7th/cmp-cmdline', 'f3fora/cmp-spell',
        },
        config = function()
            local cmp = require("cmp")
            local luasnip = require('luasnip')

            cmp.setup({
                snippet = {
                    expand = function(args)
                        luasnip.lsp_expand(args.body)
                    end
                },
                sources = {
                    { name = 'path' },
                    { name = 'nvim_lsp', keyword_length = 1 },
                    { name = 'buffer',   keyword_length = 3 },
                    { name = 'luasnip',  keyword_length = 2 },
                },
                window = {
                    completion = cmp.config.window.bordered(),
                    documentation = cmp.config.window.bordered()
                },
                formatting = {
                    fields = { 'menu', 'abbr', 'kind' },
                    format = function(entry, item)
                        local menu_icon = {
                            nvim_lsp = 'λ',
                            luasnip = '⋗',
                            buffer = 'Ω',
                            path = '🖫',
                        }

                        item.menu = menu_icon[entry.source.name]
                        return item
                    end,
                },
                mapping = {
                    ['<Up>'] = cmp.mapping.select_prev_item(select_opts),
                    ['<Down>'] = cmp.mapping.select_next_item(select_opts),

                    ['<C-u>'] = cmp.mapping.scroll_docs(-4),
                    ['<C-d>'] = cmp.mapping.scroll_docs(4),

                    ['<C-e>'] = cmp.mapping.abort(),
                    ['<C-y>'] = cmp.mapping.confirm({ select = true }),
                    ['<CR>'] = cmp.mapping.confirm({ select = false }),

                    ['<C-f>'] = cmp.mapping(function(fallback)
                        if luasnip.jumpable(1) then
                            luasnip.jump(1)
                        else
                            fallback()
                        end
                    end, { 'i', 's' }),

                    ['<C-b>'] = cmp.mapping(function(fallback)
                        if luasnip.jumpable(-1) then
                            luasnip.jump(-1)
                        else
                            fallback()
                        end
                    end, { 'i', 's' }),
                }
            })

            cmp.setup.cmdline({ '/', '?' }, {
                mapping = cmp.mapping.preset.cmdline(),
                sources = {
                    { name = 'buffer' }
                }
            })

            -- to select something in the cmdline need to use <C-n>
            cmp.setup.cmdline(':', {
                mapping = cmp.mapping.preset.cmdline(),
                sources = cmp.config.sources({
                    { name = 'path' }
                }, {
                    { name = 'cmdline' }
                }),
                matching = { disallow_symbol_nonprefix_matching = false }
            })
        end,
    }
    use {
        'nvimtools/none-ls.nvim',
        config = function()
            local null_ls = require("null-ls")
            local lsp_conf = require("lspconfig")

            -- util.root_pattern returns a function
            local find_pyproject = lsp_conf.util.root_pattern('pyproject.toml')
            local find_flake8 = lsp_conf.util.root_pattern('.flake8')
            -- local find_eslint = lsp_conf.util.root_pattern('eslint.config.js')

            local isort = null_ls.builtins.formatting.isort.with({
                cwd = function(params)
                    return find_pyproject(params.root) or params.root
                end
            })

            local black = null_ls.builtins.formatting.black.with({
                cwd = function(params)
                    return find_pyproject(params.root) or params.root
                end
            })

            local flake8 = require("none-ls.diagnostics.flake8").with({
                cwd = function(params)
                    return find_flake8(params.root) or params.root
                end
            })

            local eslint_format = require("none-ls.formatting.eslint_d").with({
                extra_filetypes = { "svelte" },
            })

            local eslint_diag = require("none-ls.diagnostics.eslint_d").with({
                extra_filetypes = { "svelte" },
            })

            local eslint_code_actions = require("none-ls.code_actions.eslint_d").with({
                extra_filetypes = { "svelte" },
            })

            local spellcheck = null_ls.builtins.completion.spell.with({
                filetypes = { "markdown" }
            })

            null_ls.setup {
                on_attach = on_attach,
                sources = {
                    eslint_format,
                    eslint_diag,
                    eslint_code_actions,
                    null_ls.builtins.code_actions.gitsigns,
                    null_ls.builtins.completion.luasnip,
                    null_ls.builtins.completion.tags,
                    spellcheck,
                    flake8,
                    isort,
                    black,
                    flake8
                }
            }
        end,
        requires = { "nvim-lua/plenary.nvim",
            "nvimtools/none-ls-extras.nvim"
        },
    }
    use {
        'folke/trouble.nvim',
        config = function()
            require("trouble").setup {
                nmap("<leader>xx", "<cmd>Trouble<cr>"),
            }
        end
    }
    use {
        "hedyhli/outline.nvim",
        config = function()
            nmap("<leader>o", "<cmd>Outline<CR>")
            require("outline").setup {
                -- Your setup opts here (leave empty to use defaults)
            }
        end,
    }
    use {
        'akinsho/bufferline.nvim',
        tag = "*",
        requires = 'nvim-tree/nvim-web-devicons',
        config = function()
            require("bufferline").setup {
                options = {
                    offsets = {
                        {
                            filetype = "NvimTree",
                            text = "Neovim - Files",
                            highlight = "Directory",
                            separator = true
                        },
                    },
                    separator_style = "slope",
                    diagnostics = "nvim_lsp",
                }
            }
        end
    }
    use {
        "nvim-lualine/lualine.nvim",
        event = "VimEnter",
        config = function()
            require("lualine").setup {
                extensions = { 'nvim-tree', 'fzf', 'symbols-outline' },
                options = {
                    component_separators = '',
                    section_separators = { left = '', right = '' },
                    theme = 'iceberg_dark',
                    globalstatus = true,
                },
                sections = {
                    -- would be nice to open code actions but w/e
                    lualine_b = {
                        {
                            'diff',
                        },
                    },
                    lualine_c = {
                        {
                            'diagnostics',
                            sources = { 'nvim_diagnostic', 'nvim_lsp' },
                            sections = { 'hint', 'warn', 'error' },
                            colored = true,
                            update_in_insert = true,
                            always_visible = false,
                            on_click = function()
                                require("trouble").toggle("diagnostics")
                            end
                        }
                    },
                    lualine_x = { 'filetype' }
                }
            }
        end,
        requires = { 'nvim-tree/nvim-web-devicons', opt = true }
    }
    use {
        "startup-nvim/startup.nvim",
        after = "nvim-web-devicons",
        requires = {
            "nvim-tree/nvim-web-devicons",
            "nvim-telescope/telescope.nvim",
            "nvim-lua/plenary.nvim"
        },
        config = function()
            local startup = require("startup")
            local headers = require("startup.headers")
            local start_fn = require("startup.functions")
            startup.setup({
                banner = {
                    type = "text",
                    align = "center",
                    title = "Header",
                    content = headers.hydra_header,
                    highlight = "Statement",
                },
                quote = {
                    type = "text",
                    align = "center",
                    fold_section = false,
                    title = "Quote",
                    margin = 5,
                    content = require("startup.functions").quote(),
                    highlight = "Constant",
                    default_color = "",
                },
                body = {
                    type = "mapping",
                    align = "center",
                    fold_section = false,
                    title = "Basic Commands",
                    margin = 5,
                    content = {
                        { " New File", "lua require'startup'.new_file()", "<leader>nf" },
                        { " Recent Files", "Telescope oldfiles", "<leader>of" },
                        { " Find File", "Telescope find_files", "<leader>ff" },
                    },
                    highlight = "String",
                    default_color = "",
                },
                plugin_info = {
                    type = "text",
                    content = require("startup.functions").packer_plugins(),
                    align = "center",
                    fold_section = false,
                    title = "",
                    margin = 5,
                    highlight = "TSString",
                    default_color = "#FFFFFF",
                },
                options = {
                    mapping_keys = true,
                },
                parts = { "banner", "quote", "body", "plugin_info" }
            })
        end
    }
    use 'RRethy/vim-illuminate'
    use {
        'nvim-telescope/telescope.nvim',
        requires = { 'nvim-lua/plenary.nvim' },
        config = function()
            require("telescope").setup {}
        end
    }
    use {
        'j-hui/fidget.nvim',
        config = function()
            require('fidget').setup {}
        end
    }
    use 'voldikss/vim-floaterm'
    use { 'norcalli/nvim-colorizer.lua',
        config = function()
            require('colorizer').setup {
                'css',
                'scss',
                'javascript',
                'toml'
            }
        end,
    }
    use { 'lewis6991/gitsigns.nvim',
        config = function()
            require('gitsigns').setup {
                on_attach = function()
                    nmap('<leader>hd', '<cmd>Gitsigns diffthis<cr>')
                    nmap('<leader>hr', '<cmd>Gitsigns reset_hunk<cr>')
                    nmap('<leader>td', '<cmd>Gitsigns toggle_deleted<cr>')
                end
            }
        end
    }
    use { 'sudormrfbin/cheatsheet.nvim',
        -- default toggle is <leader>?
        requires = {
            { 'nvim-telescope/telescope.nvim' },
            { 'nvim-lua/popup.nvim' },
            { 'nvim-lua/plenary.nvim' },
        }
    }
    use { 'stevearc/dressing.nvim' }
    if packer_bootstrap then
        require('packer').sync()
    end
    use {
        "danymat/neogen",
        config = function()
            local neogen = require('neogen')
            neogen.setup({
                snippet_engine = "luasnip",
                placeholders_hl = "None" -- fixes weird bug where the entire page gets colored with this color
            })
            nmap('<leader>d', ":lua require('neogen').generate()<CR>")
        end
    }
    use {
        "folke/which-key.nvim",
        config = function()
            local which_key = require('which-key')
            which_key.setup({
                preset = "modern"
            })
        end
        ,
    }
    use {
        'numToStr/Comment.nvim',
        config = function()
            require('Comment').setup()
        end
    }
    use({
        "kylechui/nvim-surround",
        tag = "*", -- Use for stability; omit to use `main` branch for the latest features
        config = function()
            require("nvim-surround").setup({
                -- Configuration here, or leave empty to use defaults
            })
        end
    })
    use {
        'pwntester/octo.nvim',
        requires = {
            'nvim-lua/plenary.nvim',
            'nvim-telescope/telescope.nvim',
            'nvim-tree/nvim-web-devicons',
        },
        config = function()
            require "octo".setup()
        end
    }
    use {
        "monaqa/dial.nvim",
        config = function()
            local augend = require("dial.augend")
            local dial = require("dial.config")
            local d_map = require("dial.map")

            dial.augends:register_group {
                default = {
                    augend.integer.alias.decimal,
                    augend.date.alias["%m/%d"],
                    augend.constant.alias.alpha,
                }
            }
        end
    }
    use "cocopon/iceberg.vim"
end)
